// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

namespace Smdn.Net.MessageAccessProtocols.Strings {
  public class TypedStringSet<TTypedString> :
    IReadOnlyStringSet<TTypedString>,
    IReadOnlyCollection<TTypedString>,
    ISet<TTypedString>
  {
    private class DefaultTypedStringEqualityComparer : TypedStringEqualityComparer<TTypedString> {
      protected internal override string GetStringExpressionOf(TTypedString typedString) => typedString?.ToString();
      public override bool Equals(string x, string y) => StringComparer.Ordinal.Equals(x, y);
      public override int GetHashCode(string obj) => StringComparer.Ordinal.GetHashCode(obj);
    }

    private readonly Dictionary<string, TTypedString> set;

    public bool IsReadOnly { get; }
    public int Count => set.Count;
    public TypedStringEqualityComparer<TTypedString> Comparer { get; }

    public TypedStringSet()
      : this(capacity: default,
             collection: null,
             comparer: default,
             asReadOnly: false)
    {
    }

    public TypedStringSet(TypedStringEqualityComparer<TTypedString> comparer)
      : this(capacity: default,
             collection: null,
             comparer: comparer,
             asReadOnly: false)
    {
    }

    public TypedStringSet(int capacity)
      : this(capacity: capacity,
             collection: null,
             comparer: default,
             asReadOnly: false)
    {
    }

    public TypedStringSet(int capacity, TypedStringEqualityComparer<TTypedString> comparer)
      : this(capacity: capacity,
             collection: null,
             comparer: comparer,
             asReadOnly: false)
    {
    }

    public TypedStringSet(IEnumerable<TTypedString> collection, bool asReadOnly = false)
      : this(capacity: default,
             collection: collection ?? throw new ArgumentNullException(nameof(collection)),
             comparer: default,
             asReadOnly: asReadOnly)
    {
    }

    public TypedStringSet(IEnumerable<TTypedString> collection, TypedStringEqualityComparer<TTypedString> comparer, bool asReadOnly = false)
      : this(capacity: 0,
             collection: collection ?? throw new ArgumentNullException(nameof(collection)),
             comparer: comparer,
             asReadOnly: asReadOnly)
    {
    }

    private TypedStringSet(int capacity, IEnumerable<TTypedString> collection, TypedStringEqualityComparer<TTypedString> comparer, bool asReadOnly)
    {
      IsReadOnly = asReadOnly;
      Comparer = comparer ?? new DefaultTypedStringEqualityComparer();
      set = new Dictionary<string, TTypedString>(capacity, comparer);

      if (collection != null)
        InternalAddRange(collection);
    }

    public bool Contains(string @value)
    {
      return set.ContainsKey(@value ?? throw new ArgumentNullException(nameof(@value)));
    }

    public bool TryGetValue(string @value, out TTypedString item)
    {
      return set.TryGetValue(@value ?? throw new ArgumentNullException(nameof(@value)), out item);
    }

    public void AddRange(IEnumerable<TTypedString> items)
    {
      if (items == null)
        throw new ArgumentNullException(nameof(items));

      CheckReadOnly();

      InternalAddRange(items);
    }

    private void InternalAddRange(IEnumerable<TTypedString> items)
    {
      foreach (var item in items) {
        if (item == null)
          throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(nameof(items));

        set[Comparer.GetStringExpressionOf(item)] = item;
      }
    }

    public int RemoveWhere(Predicate<TTypedString> match)
    {
      if (match == null)
        throw new ArgumentNullException(nameof(match));

      CheckReadOnly();

      var removal = this.set.Values.Where(new Func<TTypedString, bool>(match)).ToList();

      foreach (var item in removal) {
        this.set.Remove(Comparer.GetStringExpressionOf(item));
      }

      return removal.Count;
    }

#region "ISet"
    public bool Add(TTypedString item)
    {
      CheckReadOnly();

      return AddCore(item);
    }

    void ICollection<TTypedString>.Add(TTypedString item)
    {
      CheckReadOnly();

      AddCore(item);
    }

    private bool AddCore(TTypedString item)
    {
      if (item == null)
        throw new ArgumentNullException(nameof(item));

      var str = Comparer.GetStringExpressionOf(item);

      if (set.ContainsKey(str))
        return false;

      set[str] = item;

      return true;
    }

    public bool Remove(TTypedString item)
    {
      if (item == null)
        throw new ArgumentNullException(nameof(item));

      CheckReadOnly();

      return set.Remove(Comparer.GetStringExpressionOf(item));
    }

    public void Clear()
    {
      CheckReadOnly();

      set.Clear();
    }

    public void CopyTo(TTypedString[] array, int arrayIndex)
    {
      set.Values.CopyTo(array, arrayIndex);
    }

    public bool Contains(TTypedString item)
    {
      if (item == null)
        throw new ArgumentNullException(nameof(item));

      return set.ContainsKey(Comparer.GetStringExpressionOf(item));
    }

    public IEnumerator<TTypedString> GetEnumerator()
    {
      return set.Values.GetEnumerator();
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return set.Values.GetEnumerator();
    }

    private bool ToSetIf(IEnumerable<TTypedString> other, Predicate<int> countPredicate, out TypedStringSet<TTypedString> otherSet)
    {
      otherSet = null;

      var countChecked = false;

      if (countPredicate == null) {
        countChecked = true; // create set regardless of count of other collection
      }
      else {
        if (other is IReadOnlyCollection<TTypedString> readOnlyCollection) {
          if (!countPredicate(readOnlyCollection.Count))
            return false;
          countChecked = true;
        }
        else if (other is ICollection<TTypedString> collection) {
          if (!countPredicate(collection.Count))
            return false;
          countChecked = true;
        }
      }

      otherSet = other as TypedStringSet<TTypedString>;

      if (otherSet == null || !otherSet.Comparer.Equals(this.Comparer))
        otherSet = new TypedStringSet<TTypedString>(other, this.Comparer, asReadOnly: false);

      return countChecked || countPredicate(otherSet.Count);
    }

    public void ExceptWith(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      CheckReadOnly();

      foreach (var item in other) {
        this.set.Remove(Comparer.GetStringExpressionOf(item));
      }
    }

    public void IntersectWith(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      CheckReadOnly();

      if (ToSetIf(other, otherCount => 0 < otherCount, out var otherSet))
        RemoveWhere(item => !otherSet.Contains(item));
      else // if other is empty
        Clear();
    }

    public void SymmetricExceptWith(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      CheckReadOnly();

      foreach (var item in other) {
        var str = Comparer.GetStringExpressionOf(item);

        if (this.set.ContainsKey(str))
          this.set.Remove(str);
        else
          this.set.Add(str, item);
      }
    }

    public void UnionWith(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      CheckReadOnly();

      foreach (var item in other) {
        AddCore(item);
      }
    }

    public bool IsProperSubsetOf(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      if (this.set.Count == 0)
        return !other.Any();

      if (!ToSetIf(other, otherCount => this.set.Count < otherCount, out var otherSet))
        return false;

      foreach (var item in this) {
        if (!otherSet.set.ContainsKey(Comparer.GetStringExpressionOf(item)))
          return false;
      }

      return true;
    }

    public bool IsSubsetOf(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      if (!ToSetIf(other, otherCount => this.set.Count <= otherCount, out var otherSet))
        return false;

      foreach (var item in this) {
        if (!otherSet.set.ContainsKey(Comparer.GetStringExpressionOf(item)))
          return false;
      }

      return true;
    }

    public bool IsProperSupersetOf(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      if (!ToSetIf(other, otherCount => otherCount < this.set.Count, out var otherSet))
        return false;

      foreach (var item in otherSet) {
        if (!this.set.ContainsKey(Comparer.GetStringExpressionOf(item)))
          return false;
      }

      return true;
    }

    public bool IsSupersetOf(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      if (!ToSetIf(other, otherCount => otherCount <= this.set.Count, out var otherSet))
        return false;

      foreach (var item in otherSet) {
        if (!this.set.ContainsKey(Comparer.GetStringExpressionOf(item)))
          return false;
      }

      return true;
    }

    public bool Overlaps(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      foreach (var item in other) {
        if (this.set.ContainsKey(Comparer.GetStringExpressionOf(item)))
          return true;
      }

      return false;
    }

    public bool SetEquals(IEnumerable<TTypedString> other)
    {
      if (other == null)
        throw new ArgumentNullException(nameof(other));

      if (!ToSetIf(other, otherCount => this.set.Count == otherCount, out var otherSet))
        return false;

      foreach (var item in otherSet) {
        if (!this.set.ContainsKey(Comparer.GetStringExpressionOf(item)))
          return false;
      }

      return true;
    }
#endregion

    public TTypedString[] ToArray()
    {
      var array = new TTypedString[set.Count];

      set.Values.CopyTo(array, 0);

      return array;
    }

    public override string ToString() => string.Join(", ", set.Values);

    private void CheckReadOnly()
    {
      if (IsReadOnly)
        throw new NotSupportedException("set is readonly");
    }
  }
}
