using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using System.Text;
using System.Text.RegularExpressions;

using NUnit.Framework;

namespace Smdn.Net {
  public abstract class PseudoServer : IDisposable {
    protected abstract class ResponseBase {
      public string ExpectedCommand { get; }
      public string ResponseText { get; }
      public Action ActionBeforeRespond { get; }
      public Action ActionAfterRespond { get; }
      public TimeSpan ResponseDelay { get; }

      public ResponseBase(string expectedCommand,
                          string responseText,
                          Action actionBeforeRespond,
                          Action actionAfterRespond,
                          TimeSpan responseDelay)
      {
        ExpectedCommand = expectedCommand;
        ResponseText = responseText;
        ActionBeforeRespond = actionBeforeRespond;
        ActionAfterRespond = actionAfterRespond;
        ResponseDelay = responseDelay;
      }
    }

    public IPEndPoint ServerEndPoint {
      get { return server?.LocalEndpoint as IPEndPoint; }
    }

    public string Host {
      get { return ServerEndPoint.Address.ToString(); }
    }

    public int Port {
      get { return ServerEndPoint.Port; }
    }

    public string HostPort {
      get { return string.Format("{0}:{1}", Host, Port); }
    }

    public Encoding TransferCharset {
      get { return transferCharset; }
    }

    protected PseudoServer()
    {
      Initialize();
    }

    protected virtual void Initialize()
    {
      server = new TcpListener(new IPEndPoint(IPAddress.Loopback, 0));
      server.Start();

      clientRequestWaitHandle = new ManualResetEvent(false);
    }

    public void Restart()
    {
      (this as IDisposable).Dispose();

      Initialize();

      StartAcceptClient();
    }

    private void StartAcceptClient()
    {
      if (acceptClientTask != null)
        throw new InvalidOperationException("already started");

      acceptClientTask = Task.Factory.StartNew(AcceptClient);
    }

    void IDisposable.Dispose()
    {
      Exception serverProcException = null;

      if (acceptClientTask != null) {
        try {
          if (!acceptClientTask.Wait(500) && throwIfServerNotStopped)
            throw new InvalidOperationException("server not stopped");
        }
        catch (AggregateException ex) {
          serverProcException = ex.GetBaseException();
        }

        acceptClientTask = null;
      }

      server?.Stop();
      server = null;

      clientRequestWaitHandle?.Dispose();
      clientRequestWaitHandle = null;

      serverClock?.Stop();
      serverClock = null;

      if (serverProcException != null) {
        WriteLogLine("[unhandled server exception]{0}{1}", Environment.NewLine, serverProcException);

        throw serverProcException;
      }
    }

    protected void Start<TPseudoServer>(bool acceptClient,
                                        bool throwIfServerNotStopped,
                                        Action<TPseudoServer> action,
                                        [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                                        [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                                        [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
      where TPseudoServer : PseudoServer
    {
      WriteLogLine("[{0}] ({1}, {2})", callerMemberName, sourceFilePath, sourceLinerNumber);

      this.throwIfServerNotStopped = throwIfServerNotStopped;

      if (acceptClient)
        StartAcceptClient();

      try {
        action(this as TPseudoServer);
      }
      catch {
        this.throwIfServerNotStopped = false; // not to throw exception in Dispose()

        throw;
      }
    }

    public void SetCharset(Encoding newCharset)
    {
      if (newCharset == null)
        throw new ArgumentNullException("newCharset");

      transferCharset = newCharset;
    }

    public void WaitForClientRequest()
    {
      WaitForClientRequest(TimeSpan.FromSeconds(5));
    }

    public bool WaitForClientRequest(TimeSpan timeout)
    {
      clientRequestWaitHandle.Reset();

      return clientRequestWaitHandle.WaitOne(timeout);
    }

    protected void SetClientRequestReceived()
    {
      clientRequestWaitHandle.Set();
    }

    private void AcceptClient()
    {
      serverClock = Stopwatch.StartNew();

      WriteLogLine("listening...");

      using (var client = server.AcceptTcpClientAsync().Result) {
        WriteLogLine("client accepted");

        client.ReceiveTimeout = 100;

        try {
          WriteLogLine("server process started");

          ServerProc(server, client);

          WriteLogLine("server process successfully finished");
        }
        catch (Exception) {
          WriteLogLine("server process stopped by exception");

          throw;
        }

        WriteLogLine("connection closing");
      }

      WriteLogLine("connection closed");
    }

    protected abstract void ServerProc(TcpListener server, TcpClient client);

    protected void WriteLogLine(string format, params object[] args)
    {
      if (serverClock != null)
        TestContext.Out.Write("({0:D2}.{1:D4}) ", serverClock.Elapsed.Seconds, serverClock.Elapsed.Milliseconds);

      TestContext.Out.WriteLine(format, args);
    }

    protected void WriteLog(string line)
    {
      if (serverClock != null)
        TestContext.Out.Write("({0:D2}.{1:D4}) ", serverClock.Elapsed.Seconds, serverClock.Elapsed.Milliseconds);

      TestContext.Out.Write(line);
    }

    private static string[] logLineSplitter = new[] {"\r\n"};

    protected void WriteLog(string prefix, string line)
    {
      if (serverClock != null)
        TestContext.Out.Write("({0:D2}.{1:D4}) ", serverClock.Elapsed.Seconds, serverClock.Elapsed.Milliseconds);

      var firstLine = true;
      var indent = new string(' ', 2 + 1 + 4 + 1 + prefix.Length) + "\t";

      foreach (var l in line.TrimEnd().Split(logLineSplitter, StringSplitOptions.None)) {
        if (firstLine) {
          TestContext.Out.Write(prefix + "\t");
          firstLine = false;
        }
        else {
          TestContext.Out.Write(indent);
        }

        TestContext.Out.WriteLine(l);
      }
    }

    private static readonly byte[] peekClientBuffer = new byte[1];

    protected bool IsConnected(TcpClient client)
    {
      try {
        if (client.Client.Poll(0, SelectMode.SelectRead)) {
          if (client.Client.Receive(peekClientBuffer, SocketFlags.Peek) == 0) {
            WriteLogLine("client disconnected");
            return false;
          }
        }
      }
      catch (SocketException ex) {
        switch (ex.SocketErrorCode) {
          case SocketError.ConnectionAborted:
          case SocketError.ConnectionReset:
            return false;

          default:
            TestContext.Out.WriteLine("unexpected socket error: {0}", ex.SocketErrorCode);
            throw;
        }
      }

      return true;
    }

    public class UnexpectedCommandException : InvalidOperationException {
      public UnexpectedCommandException(string expectedCommand, string actualCommand)
        : base(string.Format("expected '{0}' but was '{1}'",
                             expectedCommand.Replace("\r", "\\r").Replace("\n", "\\n"),
                             actualCommand?.Replace("\r", "\\r")?.Replace("\n", "\\n")))
      {
      }
    }

    protected void Send(ResponseBase response, Stream requestStream, string command = null)
    {
      if (!string.IsNullOrEmpty(response.ExpectedCommand)) {
        if (string.IsNullOrEmpty(command))
          throw new UnexpectedCommandException(response.ExpectedCommand, command);

        if (0 > command.IndexOf(response.ExpectedCommand, StringComparison.OrdinalIgnoreCase))
          throw new UnexpectedCommandException(response.ExpectedCommand, command);
      }

      response.ActionBeforeRespond?.Invoke();

      var bytes = transferCharset.GetBytes(response.ResponseText);

      Thread.Sleep(response.ResponseDelay);

      requestStream.Write(bytes, 0, bytes.Length);

      WriteLog("S:", EscapeLogControlChars(response.ResponseText));

      response.ActionAfterRespond?.Invoke();
    }

    private static readonly Regex regexEscapeChars = new Regex(@"[\x00-\x09\x0b\x0c\x0e-\x1f\x7f]+|\r(?!\n)|(?<!\r)\n");

    protected string EscapeLogControlChars(string str)
    {
      str = regexEscapeChars.Replace(str, match => {
        var sb = new StringBuilder(match.Value.Length * 4);

        foreach (var c in match.Value) {
          switch (c) {
            case '\n': sb.Append("\\n"); break;
            case '\r': sb.Append("\\r"); break;
            case '\t': sb.Append("\\t"); break;
            default:
              sb.Append("\\x");
              sb.Append(((int)c).ToString("X2"));
              break;
          }
        }

        return sb.ToString();
      });

      return str;
    }

    private Stopwatch serverClock;
    private TcpListener server;
    private Task acceptClientTask;
    private ManualResetEvent clientRequestWaitHandle;
    private bool throwIfServerNotStopped = true;
    private Encoding transferCharset = Smdn.Text.Encodings.OctetEncoding.EightBits;
  }
}
