// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Connection;
using Smdn.Net.MessageAccessProtocols.Strings;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Session;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Pop3.Client {
  public partial class PopClient : IDisposable {
    internal const string ObsoleteAttributeMessage_UseOptionalParameterOverloads = "省略可能な引数を取るバージョンのメソッドオーバーロードを使用してください/Use method version that takes optional parameters";

    public PopClientProfile Profile { get; }

    public int Timeout {
      get { return (session == null) ? Profile.Timeout : session.TransactionTimeout; }
      set
      {
        Profile.Timeout = value;

        if (session != null)
          session.TransactionTimeout = value;
      }
    }

    public int SendTimeout {
      get { return (session == null) ? Profile.SendTimeout : session.SendTimeout; }
      set
      {
        Profile.SendTimeout = value;

        if (session != null)
          session.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get { return (session == null) ? Profile.ReceiveTimeout : session.ReceiveTimeout; }
      set
      {
        Profile.ReceiveTimeout = value;

        if (session != null)
          session.ReceiveTimeout = value;
      }
    }

    public bool IsConnected {
      get {
        if (session == null)
          return false;

        switch (session.State) {
          case PopSessionState.NotConnected:
          case PopSessionState.Disconnected:
            return false;
          default:
            return true;
        }
      }
    }

    public bool IsBusy {
      get { return sessionStatus == SessionStatus_Busy; }
    }

    internal PopSession Session {
      get { ThrowIfNotConnected(); return session; }
    }

    public bool IsSecureSession {
      get { ThrowIfNotConnected(); return session.IsSecureConnection; }
    }

    public IReadOnlyStringSet<PopCapability> ServerCapabilities {
      get { ThrowIfNotConnected(); return session.ServerCapabilities; }
    }

    /*
     * construction
     */
    public PopClient()
      : this(new PopClientProfile())
    {
    }

    public PopClient(Uri authority)
      : this(new PopClientProfile(authority))
    {
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public PopClient(string host, string userName)
      : this(new PopClientProfile(host, userName: userName))
    {
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public PopClient(string host, bool securePort, string userName)
      : this(new PopClientProfile(host, securePort: securePort, userName: userName))
    {
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public PopClient(string host, int port, string userName)
      : this(new PopClientProfile(host, port: port, userName: userName))
    {
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public PopClient(string host, int port, string userName, string authType)
      : this(new PopClientProfile(host, port: port, userName: userName, authType: authType))
    {
    }

    public PopClient(string host,
                     int port = PopClientProfile.DefaultPort,
                     bool securePort = PopClientProfile.DefaultSecurePort,
                     string userName = PopClientProfile.DefaultUserName,
                     string authType = PopClientProfile.DefaultAuthType,
                     int timeout = PopClientProfile.DefaultTimeout,
                     int sendTimeout = PopClientProfile.DefaultSendTimeout,
                     int receiveTimeout = PopClientProfile.DefaultReceiveTimeout,
                     bool useTlsIfAvailable = PopClientProfile.DefaultUseTlsIfAvailable,
                     IEnumerable<string> usingSaslMechanisms = null,
                     bool allowInsecureLogin = PopClientProfile.DefaultAllowInsecureLogin,
                     bool allowStandardLogin = PopClientProfile.DefaultAllowStandardLogin,
                     bool allowAllowApopLogin = PopClientProfile.DefaultAllowApopLogin)
      : this(new PopClientProfile(host,
                                  port: port,
                                  securePort: securePort,
                                  userName: userName,
                                  authType: authType,
                                  timeout: timeout,
                                  sendTimeout: sendTimeout,
                                  receiveTimeout: receiveTimeout,
                                  useTlsIfAvailable: useTlsIfAvailable,
                                  usingSaslMechanisms: usingSaslMechanisms,
                                  allowInsecureLogin: allowInsecureLogin,
                                  allowStandardLogin: allowStandardLogin,
                                  allowApopLogin: allowAllowApopLogin))
    {
    }

    public PopClient(PopClientProfile profile)
    {
      Profile = profile ?? throw new ArgumentNullException(nameof(profile));
    }

    /*
     * connect impl
     */
    private static async Task<PopSession> CreateSession(ConnectParams @params)
    {
      try {
        return await PopSession.CreateAsync(
          profile: @params.Profile,
          authMechanism: @params.AuthMechanism,
          resolveHostAddress: null,
          createSslStream: @params.CreateSslStream ?? PopSslConnection.CreateSslStream,
          cancellationToken: @params.CancellationToken
        ).ConfigureAwait(false);
      }
      finally {
        @params.Profile.Credentials = null;
      }
    }

    private sealed class ConnectParams {
      public readonly PopClientProfile Profile;
      public readonly SaslClientMechanism AuthMechanism;
      public readonly UpgradeConnectionStreamCallback CreateSslStream;
      public readonly CancellationToken CancellationToken;

      public ConnectParams(PopClientProfile profile,
                           ICredentialsByHost credentials,
                           UpgradeConnectionStreamCallback createSslStream,
                           CancellationToken cancellationToken)
      {
        Profile = profile.Clone();
        Profile.Credentials = credentials ?? throw new ArgumentNullException(nameof(credentials));

        CreateSslStream = createSslStream;
        CancellationToken = cancellationToken;
      }

      public ConnectParams(PopClientProfile profile,
                           SaslClientMechanism authMechanism,
                           UpgradeConnectionStreamCallback createSslStream,
                           CancellationToken cancellationToken)
      {
        Profile = profile.Clone();
        AuthMechanism = authMechanism ?? throw new ArgumentNullException(nameof(authMechanism));
        CreateSslStream = createSslStream;
        CancellationToken = cancellationToken;
      }
    }

    /*
     * BeginConnect()/EndConnect()
     */
    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public IAsyncResult BeginConnect(string password,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
      => BeginConnect(password, null, asyncCallback, asyncState);

    public IAsyncResult BeginConnect(string password,
                                     UpgradeConnectionStreamCallback createSslStream = null,
                                     AsyncCallback asyncCallback = null,
                                     object asyncState = null)
    {
      return BeginConnect(new ConnectParams(Profile,
                                            new NetworkCredential(Profile.UserName, password),
                                            createSslStream,
                                            CancellationToken.None),
                          asyncCallback,
                          asyncState);
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
      => BeginConnect(credentials, null, asyncCallback, asyncState);

    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     UpgradeConnectionStreamCallback createSslStream = null,
                                     AsyncCallback asyncCallback = null,
                                     object asyncState = null)
    {
      return BeginConnect(new ConnectParams(Profile,
                                            credentials,
                                            createSslStream,
                                            CancellationToken.None),
                          asyncCallback,
                          asyncState);
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
      => BeginConnect(authMechanism, null, asyncCallback, asyncState);

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     UpgradeConnectionStreamCallback createSslStream = null,
                                     AsyncCallback asyncCallback = null,
                                     object asyncState = null)
    {
      return BeginConnect(new ConnectParams(Profile,
                                            authMechanism,
                                            createSslStream,
                                            CancellationToken.None),
                          asyncCallback,
                          asyncState);
    }

    private Task beginEndConnectTask = null;

    private IAsyncResult BeginConnect(ConnectParams @params,
                                      AsyncCallback asyncCallback,
                                      object asyncState)
    {
      var task = ConnectAsync(@params);
      var tcs = new TaskCompletionSource<object>(asyncState);

      task.ContinueWith(t => {
        if (t.IsFaulted)
          tcs.TrySetException(t.Exception.InnerExceptions);
        else if (t.IsCanceled)
          tcs.TrySetCanceled();
        else
          tcs.TrySetResult(null);

        asyncCallback?.Invoke(tcs.Task);
      }, TaskScheduler.Default);

      beginEndConnectTask = tcs.Task;

      return beginEndConnectTask;
    }

    public void EndConnect(IAsyncResult asyncResult)
    {
      if (asyncResult == null)
        throw new ArgumentNullException(nameof(asyncResult));
      if (asyncResult != beginEndConnectTask)
        throw ExceptionUtils.CreateArgumentMustBeValidIAsyncResult(nameof(asyncResult));

      try {
        beginEndConnectTask.Wait();
      }
      finally {
        beginEndConnectTask = null;
      }
    }

    /*
     * ConnectAsync()
     */
    public Task ConnectAsync(string password,
                             UpgradeConnectionStreamCallback createSslStream = null,
                             CancellationToken cancellationToken = default)
    {
      return ConnectAsync(new ConnectParams(Profile,
                                            new NetworkCredential(Profile.UserName, password),
                                            createSslStream,
                                            cancellationToken));
    }

    public Task ConnectAsync(ICredentialsByHost credentials,
                             UpgradeConnectionStreamCallback createSslStream = null,
                             CancellationToken cancellationToken = default)
    {
      return ConnectAsync(new ConnectParams(Profile,
                                            credentials,
                                            createSslStream,
                                            cancellationToken));
    }

    public Task ConnectAsync(SaslClientMechanism authMechanism,
                             UpgradeConnectionStreamCallback createSslStream = null,
                             CancellationToken cancellationToken = default)
    {
      return ConnectAsync(new ConnectParams(Profile,
                                            authMechanism,
                                            createSslStream,
                                            cancellationToken));
    }

    private Task ConnectAsync(ConnectParams @params)
    {
      ThrowIfAlreadyConnected();

      byDisposeCancellationTokenSource?.Dispose(); // dispose previously used CancellationTokenSource

      byDisposeCancellationTokenSource = new CancellationTokenSource();

      return RunExclusivelyAsync(async () => {
        PopSession createdSession = null;

        try {
          createdSession = await CreateSession(@params).ConfigureAwait(false);
        }
        finally {
          if (byDisposeCancellationTokenSource.IsCancellationRequested)
            createdSession?.Dispose();
          else
            this.session = createdSession;
        }
      }, throwIfNotConnected: false);
    }

    /*
     * Connect()
     */
    public void Connect(string password,
                        UpgradeConnectionStreamCallback createSslStream = null,
                        CancellationToken cancellationToken = default)
    {
      Connect(new ConnectParams(Profile,
                                new NetworkCredential(Profile.UserName, password),
                                createSslStream,
                                cancellationToken));
    }

    public void Connect(ICredentialsByHost credentials,
                        UpgradeConnectionStreamCallback createSslStream = null,
                        CancellationToken cancellationToken = default)
    {
      Connect(new ConnectParams(Profile,
                                credentials,
                                createSslStream,
                                cancellationToken));
    }

    public void Connect(SaslClientMechanism authMechanism,
                        UpgradeConnectionStreamCallback createSslStream = null,
                        CancellationToken cancellationToken = default)
    {
      Connect(new ConnectParams(Profile,
                                authMechanism,
                                createSslStream,
                                cancellationToken));
    }

    private void Connect(ConnectParams @params)
    {
      ThrowIfAlreadyConnected();

      RunExclusively(() => {
        session = CreateSession(@params).GetAwaiter().GetResult();
      }, false);
    }

    /*
     * destruction
     */
    void IDisposable.Dispose()
    {
      DetachFromSession();

      session?.Dispose();
      session = null;

      byDisposeCancellationTokenSource?.Cancel(); // cancel and dispose async-connected session if created
    }

    private void DetachFromSession()
    {
      lock (messages) {
        foreach (var message in messages.Values) {
          message.Client = null;
        }

        messages.Clear();
      }
    }

    public void Logout() => Disconnect(logout: true);

    public void Disconnect(bool logout = false)
    {
      RunExclusively(() => {
        if (session == null)
          return;

        DetachFromSession();

        session.Disconnect(logout);
        session = null;
      }, false);
    }

    /*
     * methods for exclusive execution
     */
    private const int SessionStatus_Ready = 0;
    private const int SessionStatus_Busy = 1;

    private int sessionStatus = SessionStatus_Ready;

    private void SetBusy()
    {
      if (SessionStatus_Busy == Interlocked.CompareExchange(ref sessionStatus, SessionStatus_Busy, SessionStatus_Ready))
        throw new InvalidOperationException("client is busy; concurrent operations are not supported");
    }

    private void UnsetBusy()
    {
      Interlocked.Exchange(ref sessionStatus, SessionStatus_Ready);
    }

    internal void RunExclusively(Action operation, bool throwIfNotConnected = true)
    {
      if (throwIfNotConnected)
        ThrowIfNotConnected();

      SetBusy();

      try {
        operation();
      }
      finally {
        UnsetBusy();
      }
    }

    internal T RunExclusively<T>(Func<T> operation)
    {
      ThrowIfNotConnected();

      SetBusy();

      try {
        return operation();
      }
      finally {
        UnsetBusy();
      }
    }

    internal Task RunExclusivelyAsync(Func<Task> operation, bool throwIfNotConnected = true)
    {
      if (throwIfNotConnected)
        ThrowIfNotConnected();

      SetBusy();

      return Run();

      async Task Run()
      {
        try {
          await operation().ConfigureAwait(false);
        }
        finally {
          UnsetBusy();
        }
      }
    }

    internal Task StartTaskExclusively(Action operation,
                                       CancellationToken cancellationToken,
                                       bool throwIfNotConnected = true)
    {
      if (throwIfNotConnected)
        ThrowIfNotConnected();

      SetBusy();

      var t = Task.Factory.StartNew(() => {
        try {
          operation();
        }
        finally {
          UnsetBusy();
        }
      }, cancellationToken);

      // XXX
      if (cancellationToken.IsCancellationRequested && t.Status == TaskStatus.WaitingToRun)
        UnsetBusy();
      else if (t.IsCanceled)
        UnsetBusy();

      return t;
    }

    internal Task<T> StartTaskExclusively<T>(Func<T> operation,
                                             CancellationToken cancellationToken)
    {
      ThrowIfNotConnected();

      SetBusy();

      var t = Task.Factory.StartNew(() => {
        try {
          return operation();
        }
        finally {
          UnsetBusy();
        }
      }, cancellationToken);

      // XXX
      if (cancellationToken.IsCancellationRequested && t.Status == TaskStatus.WaitingToRun)
        UnsetBusy();
      else if (t.IsCanceled)
        UnsetBusy();

      return t;
    }

    /*
     * methods for checking and validating internal status
     */
    private void ThrowIfAlreadyConnected()
    {
      if (IsConnected)
        throw new InvalidOperationException("already connected");
    }

    private void ThrowIfNotConnected()
    {
      if (!IsConnected)
        throw new InvalidOperationException("not connected or already disconnected");
    }

#if DEBUG
    internal void ThrowIfNotInExclusiveOperation()
    {
      if (sessionStatus == SessionStatus_Ready)
        throw new InvalidOperationException("not in exclusive operation");
    }
#endif

    public override string ToString()
    {
      return string.Format("{{PopClient: Authority='{0}', IsConnected={1}}}",
                           Profile.Authority,
                           IsConnected);
    }

    private PopSession session;
    private CancellationTokenSource byDisposeCancellationTokenSource;
  }
}
