// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Session;
using Smdn.Net.Pop3.Urls;

namespace Smdn.Net.Pop3.Client {
#if SERIALIZATION
  [Serializable]
#endif
  public class PopClientProfile :
    IPopSessionProfile
#if SERIALIZATION
    , ISerializable
#endif
#if NETFRAMEWORK || NETSTANDARD2_0
    , ICloneable
#endif
  {
    public Uri Authority => authority.Uri;

    public string Host {
      get => authority.Host;
      set {
        if (value == null)
          throw new ArgumentNullException(nameof(Host));
        if (value.Length == 0)
          throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(Host));

        authority.Host = value;
      }
    }

    public int Port {
      get => authority.Port;
      set => authority.Port = value;
    }

    public bool SecurePort {
      get => string.Equals(authority.Scheme, PopUri.UriSchemePops, StringComparison.OrdinalIgnoreCase);
      set => authority.Scheme = value ? PopUri.UriSchemePops : PopUri.UriSchemePop;
    }

    public string UserName {
      get => authority.UserName;
      set => authority.UserName = value;
    }

    public string AuthType {
      get => authority.AuthType?.Value;
      set => authority.AuthType = value == null ? null : PopAuthenticationMechanism.GetKnownOrCreate(value);
    }

    private int timeout = DefaultTimeout;

    public int Timeout {
      get { return timeout; }
      set
      {
        if (value < -1)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(Timeout), value);
        timeout = value;
      }
    }

    private int sendTimeout = DefaultSendTimeout;

    public int SendTimeout {
      get { return sendTimeout; }
      set
      {
        if (value < -1)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(SendTimeout), value);
        sendTimeout = value;
      }
    }

    private int receiveTimeout = DefaultReceiveTimeout;

    public int ReceiveTimeout {
      get { return receiveTimeout; }
      set
      {
        if (value < -1)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(ReceiveTimeout), value);
        receiveTimeout = value;
      }
    }

#if SERIALIZATION
    [field: NonSerialized]
#endif
    internal ICredentialsByHost Credentials { get; set; } = null;
    ICredentialsByHost IPopSessionProfile.Credentials => Credentials;

    public bool UseTlsIfAvailable { get; set; } = DefaultUseTlsIfAvailable;

    public IList<string> UsingSaslMechanisms { get; private set; } = new List<string>(DefaultUsingSaslMechanisms);
    IReadOnlyList<string> IPopSessionProfile.UsingSaslMechanisms => UsingSaslMechanisms as List<string>;

    public bool AllowInsecureLogin { get; set; } = DefaultAllowInsecureLogin;
    public bool AllowStandardLogin { get; set; } = DefaultAllowStandardLogin;
    public bool AllowApopLogin { get; set; } = DefaultAllowApopLogin;

    /*
     * constructors
     */
    public PopClientProfile()
    {
      this.authority = new PopUriBuilder();
    }

    public PopClientProfile(Uri authority)
    {
      this.authority = new PopUriBuilder(authority);
    }

    internal const int DefaultPort = -1;
    internal const bool DefaultSecurePort = false;
    internal const string DefaultUserName = null;
    internal const string DefaultAuthType = null;
    internal const int DefaultTimeout = System.Threading.Timeout.Infinite;
    internal const int DefaultSendTimeout = System.Threading.Timeout.Infinite;
    internal const int DefaultReceiveTimeout = System.Threading.Timeout.Infinite;
    internal const bool DefaultUseTlsIfAvailable = true;
    internal static readonly string[] DefaultUsingSaslMechanisms = { "DIGEST-MD5", "CRAM-MD5" };
    internal const bool DefaultAllowInsecureLogin = false;
    internal const bool DefaultAllowStandardLogin = true;
    internal const bool DefaultAllowApopLogin = true;

    public PopClientProfile(string host,
                            int port = DefaultPort,
                            bool securePort = DefaultSecurePort,
                            string userName = DefaultUserName,
                            string authType = DefaultAuthType,
                            int timeout = DefaultTimeout,
                            int sendTimeout = DefaultSendTimeout,
                            int receiveTimeout = DefaultReceiveTimeout,
                            bool useTlsIfAvailable = DefaultUseTlsIfAvailable,
                            IEnumerable<string> usingSaslMechanisms = null,
                            bool allowInsecureLogin = DefaultAllowInsecureLogin,
                            bool allowStandardLogin = DefaultAllowStandardLogin,
                            bool allowApopLogin = DefaultAllowApopLogin)
      : this()
    {
      Host = host;
      Port = port;
      SecurePort = securePort;
      UserName = userName;
      AuthType = authType;
      Timeout = timeout;
      SendTimeout = sendTimeout;
      ReceiveTimeout = receiveTimeout;
      UseTlsIfAvailable = useTlsIfAvailable;
      SetUsingSaslMechanisms(usingSaslMechanisms ?? DefaultUsingSaslMechanisms);
      AllowInsecureLogin = allowInsecureLogin;
      AllowStandardLogin = allowStandardLogin;
      AllowApopLogin = allowApopLogin;
    }

#if SERIALIZATION
    protected PopClientProfile(SerializationInfo info, StreamingContext context)
    {
      authority = new PopUriBuilder(info.GetString(nameof(Authority)));
      Timeout = info.GetInt32(nameof(Timeout));
      SendTimeout = info.GetInt32(nameof(SendTimeout));
      ReceiveTimeout = info.GetInt32(nameof(ReceiveTimeout));
      UseTlsIfAvailable = info.GetBoolean(nameof(UseTlsIfAvailable));
      UsingSaslMechanisms = new List<string>((string[])info.GetValue(nameof(UsingSaslMechanisms), typeof(string[])));
      AllowInsecureLogin = info.GetBoolean(nameof(AllowInsecureLogin));
      AllowStandardLogin = info.GetBoolean(nameof(AllowStandardLogin));
      AllowApopLogin = info.GetBoolean(nameof(AllowApopLogin));
    }

    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue(nameof(Authority), Authority.ToString());
      info.AddValue(nameof(Timeout), Timeout);
      info.AddValue(nameof(SendTimeout), SendTimeout);
      info.AddValue(nameof(ReceiveTimeout), ReceiveTimeout);
      info.AddValue(nameof(UseTlsIfAvailable), UseTlsIfAvailable);
      info.AddValue(nameof(UsingSaslMechanisms), UsingSaslMechanisms.ToArray());
      info.AddValue(nameof(AllowInsecureLogin), AllowInsecureLogin);
      info.AddValue(nameof(AllowStandardLogin), AllowStandardLogin);
      info.AddValue(nameof(AllowApopLogin), AllowApopLogin);
    }
#endif

    public void SetUsingSaslMechanisms(IEnumerable<string> mechanisms)
    {
      if (mechanisms == null)
        throw new ArgumentNullException(nameof(mechanisms));

      UsingSaslMechanisms.Clear();

      foreach (var mechanism in mechanisms) {
        UsingSaslMechanisms.Add(mechanism);
      }
    }

    public PopClientProfile Clone()
    {
      var cloned = (PopClientProfile)MemberwiseClone();

      cloned.authority = this.authority.Clone();

      cloned.UsingSaslMechanisms = new List<string>(this.UsingSaslMechanisms);

      return cloned;
    }

#if NETFRAMEWORK || NETSTANDARD2_0
    object ICloneable.Clone()
    {
      return Clone();
    }
#endif

    private PopUriBuilder authority;
  }
}
