// 
// Copyright (c) 2019 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.IO;
using Smdn.IO.Streams;

using ReportRetrieveProgress = Smdn.Net.Pop3.Protocol.Transmission.Transaction.RetrTopTransactionBase.ReportRetrieveProgress;

namespace Smdn.Net.Pop3.Client {
  partial class PopMessage {
    private static readonly Encoding DefaultCharset = Encoding.GetEncoding("ISO-8859-1");

    internal MessageBody Body { get; }

    internal class MessageBody {
      public PopMessage Message { get; }

      public MessageBody(PopMessage message)
      {
        this.Message = message;
      }

      private Stream Retrieve(int maxLines,
                              Func<Stream> openRetrieveTargetStream,
                              CancellationToken cancellationToken,
                              IProgress<PopMessageRetrieveProgressInfo> progress)
      {
        var retrieveTargetStream = openRetrieveTargetStream?.Invoke();
        var closeStreamIfCancellationRequested = (openRetrieveTargetStream == null);

        var reportProgress = (progress == null)
          ? (ReportRetrieveProgress)null
          : (ReportRetrieveProgress)((octets, lines) => progress.Report(new PopMessageRetrieveProgressInfo(octets, lines, Message.ScanList.SizeInOctets, cancellationToken)));

        if (maxLines == EntireMessage)
          PopClient.ThrowIfError(Message.Client.Session.Retr(Message.MessageNumber, reportProgress, ref retrieveTargetStream));
        else
          PopClient.ThrowIfError(Message.Client.Session.Top(Message.MessageNumber, maxLines, reportProgress, ref retrieveTargetStream));

        if (Message.Client.DeleteAfterRetrieve) {
          // check cancellation requested before marking as deleted
          if (cancellationToken.IsCancellationRequested && closeStreamIfCancellationRequested)
            retrieveTargetStream.Dispose();

          cancellationToken.ThrowIfCancellationRequested();

          if (!Message.isMarkedAsDeleted)
            Message.MarkAsDeletedCore();
        }

        if (cancellationToken.IsCancellationRequested && closeStreamIfCancellationRequested)
          // check cancellation requested before yielding result
          retrieveTargetStream.Dispose();

        cancellationToken.ThrowIfCancellationRequested();

        return retrieveTargetStream;
      }

      private TResult ReadMessageBody<TResult, TArg>(
        Func<Stream, TArg, TResult> messageStreamConverter,
        TArg arg,
        int maxLines,
        CancellationToken cancellationToken,
        IProgress<PopMessageRetrieveProgressInfo> progress,
        bool keepStreamOpen = false,
        Func<Stream> openRetrieveTargetStream = null
      )
      {
        Message.ThrowIfSessionClosed();
        Message.ThrowIfMarkedAsDeleted();
        Message.ThrowIfMaxLinesOutOfRange(maxLines);

        var messageBodyStream = Message.Client.RunExclusively(() => Retrieve(maxLines, openRetrieveTargetStream, cancellationToken, progress));

        if (keepStreamOpen) {
          return messageStreamConverter(messageBodyStream, arg);
        }
        else {
          using (messageBodyStream) {
            return messageStreamConverter(messageBodyStream, arg);
          }
        }
      }

      private Task<TResult> ReadMessageBodyAsync<TResult, TArg>(
        Func<Stream, TArg, Task<TResult>> messageStreamConverter,
        TArg arg,
        int maxLines,
        CancellationToken cancellationToken,
        IProgress<PopMessageRetrieveProgressInfo> progress,
        bool keepStreamOpen = false,
        Func<Stream> openRetrieveTargetStream = null
      )
      {
        Message.ThrowIfSessionClosed();
        Message.ThrowIfMarkedAsDeleted();
        Message.ThrowIfMaxLinesOutOfRange(maxLines);

        return Read();

        async Task<TResult> Read()
        {
          var messageBodyStream = await Message.Client.StartTaskExclusively(() => Retrieve(maxLines, openRetrieveTargetStream, cancellationToken, progress), cancellationToken).ConfigureAwait(false);

          if (keepStreamOpen) {
            return await messageStreamConverter(messageBodyStream, arg).ConfigureAwait(false);
          }
          else {
            using (messageBodyStream) {
              return await messageStreamConverter(messageBodyStream, arg).ConfigureAwait(false);
            }
          }
        }
      }

      /*
       * OpenRead()
       */
      private static Stream NullConverter(Stream stream, int _) => stream;
      private static Task<Stream> NullAsyncConverter(Stream stream, int _) => Task.FromResult(stream);

      public Stream OpenRead(
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: NullConverter,
        arg: default(int),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress,
        keepStreamOpen: true
      );

      public Task<Stream> OpenReadAsync(
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: NullAsyncConverter,
        arg: default(int),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress,
        keepStreamOpen: true
      );

      /*
       * ReadAs<TResult>(Converter<Stream, TResult>)
       */
      private static TResult ReadAsConverter<TResult>(Stream stream, Converter<Stream, TResult> converter)
        => converter(stream);

      public TResult ReadAs<TResult>(
        Converter<Stream, TResult> converter,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: ReadAsConverter,
        arg: converter ?? throw new ArgumentNullException(nameof(converter)),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task<TResult> ReadAsAsync<TResult>(
        Converter<Stream, Task<TResult>> converter,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: ReadAsConverter,
        arg: converter ?? throw new ArgumentNullException(nameof(converter)),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      /*
       * ReadAs<TResult>(Func<Stream, TArg, TResult>)
       */
      private static TResult ReadAsConverter<TArg, TResult>(
        Stream stream,
        (Func<Stream, TArg, TResult> converter, TArg arg) args
      ) => args.converter(stream, args.arg);

      public TResult ReadAs<TArg, TResult>(
        Func<Stream, TArg, TResult> converter,
        TArg arg,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: ReadAsConverter,
        arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task<TResult> ReadAsAsync<TArg, TResult>(
        Func<Stream, TArg, Task<TResult>> converter,
        TArg arg,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: ReadAsConverter,
        arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      /*
       * ReadAsByteArray()
       */
      private static byte[] ReadAsByteArrayConverter(Stream stream) => stream.ReadToEnd();
      private static Task<byte[]> ReadAsByteArrayAsyncConverter(Stream stream) => stream.ReadToEndAsync();

      public byte[] ReadAsByteArray(
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadAs(
        converter: ReadAsByteArrayConverter,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task<byte[]> ReadAsByteArrayAsync(
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadAsAsync(
        converter: ReadAsByteArrayAsyncConverter,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      /*
       * WriteTo(Stream)
       */
      private static void ThrowIfStreamIsNullOrNotWritable(Stream stream)
      {
        if (stream == null)
          throw new ArgumentNullException(nameof(stream));
        if (!stream.CanWrite)
          throw ExceptionUtils.CreateArgumentMustBeWritableStream(nameof(stream));
      }

      private static int WriteToConverter(Stream stream, int _) => _;
      private static Task<int> WriteToAsyncConverter(Stream stream, int _) => Task.FromResult(_);

      public void WriteTo(
        Stream stream,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      )
      {
        ThrowIfStreamIsNullOrNotWritable(stream);

        ReadMessageBody(
          messageStreamConverter: WriteToConverter,
          arg: default(int),
          maxLines: maxLines,
          cancellationToken: cancellationToken,
          progress: progress,
          keepStreamOpen: true,
          openRetrieveTargetStream: () => stream
        );
      }

      public Task WriteToAsync(
        Stream stream,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      )
      {
        ThrowIfStreamIsNullOrNotWritable(stream);

        return ReadMessageBodyAsync(
          messageStreamConverter: WriteToAsyncConverter,
          arg: default(int),
          maxLines: maxLines,
          cancellationToken: cancellationToken,
          progress: progress,
          keepStreamOpen: true,
          openRetrieveTargetStream: () => stream
        );
      }

      /*
       * WriteTo(BinaryWriter)
       */
      private static Stream WriteToConverter(Stream stream, BinaryWriter writer)
      {
        if (stream is ChunkedMemoryStream chunkedStream)
          stream.CopyTo(writer, chunkedStream.ChunkSize);
        else
          stream.CopyTo(writer);

        return stream;
      }

      private static async Task<Stream> WriteToAsyncConverter(Stream stream, BinaryWriter writer)
      {
        const int defaultBufferSize = 8 * 1024;

        if (stream is ChunkedMemoryStream chunkedStream)
          await stream.CopyToAsync(writer, chunkedStream.ChunkSize).ConfigureAwait(false);
        else
          await stream.CopyToAsync(writer, defaultBufferSize).ConfigureAwait(false);

        return stream;
      }

      public void WriteTo(
        BinaryWriter writer,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: WriteToConverter,
        arg: writer ?? throw new ArgumentNullException(nameof(writer)),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task WriteToAsync(
        BinaryWriter writer,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: WriteToAsyncConverter,
        arg: writer ?? throw new ArgumentNullException(nameof(writer)),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      /*
       * Save()
       */
      public void Save(
        string path,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      Save(
        file: new FileInfo(path),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      internal void Save(
        FileInfo file,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: NullConverter,
        arg: default(int),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress,
        openRetrieveTargetStream: () => file.Open(FileMode.Create, FileAccess.Write)
      );

      public Task SaveAsync(
        string path,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      SaveAsync(
        file: new FileInfo(path),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      internal Task SaveAsync(
        FileInfo file,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: NullAsyncConverter,
        arg: default(int),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress,
        openRetrieveTargetStream: () => file.Open(FileMode.Create, FileAccess.Write)
      );

      /*
       * OpenText()
       */
      private static StreamReader OpenTextConverter(Stream stream, Encoding charset) => new StreamReader(stream, charset);
      private static Task<StreamReader> OpenTextAsyncConverter(Stream stream, Encoding charset) => Task.FromResult(new StreamReader(stream, charset));

      public StreamReader OpenText(
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: OpenTextConverter,
        arg: charset ?? DefaultCharset,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress,
        keepStreamOpen: true
      );

      public Task<StreamReader> OpenTextAsync(
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: OpenTextAsyncConverter,
        arg: charset ?? DefaultCharset,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress,
        keepStreamOpen: true
      );

      /*
       * ReadAs<TResult>(Converter<StreamReader, TResult>)
       */
      private static TResult ReadAsConverter<TResult>(
        Stream stream,
        (Converter<StreamReader, TResult> converter, Encoding charset) args
      ) => args.converter(new StreamReader(stream, args.charset));

      public TResult ReadAs<TResult>(
        Converter<StreamReader, TResult> converter,
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: ReadAsConverter,
        arg: (converter ?? throw new ArgumentNullException(nameof(converter)), charset ?? DefaultCharset),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task<TResult> ReadAsAsync<TResult>(
        Converter<StreamReader, Task<TResult>> converter,
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: ReadAsConverter,
        arg: (converter ?? throw new ArgumentNullException(nameof(converter)), charset ?? DefaultCharset),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      /*
       * ReadAs<TResult>(Func<StreamReader, TArg, TResult>)
       */
      private static TResult ReadAsConverter<TArg, TResult>(
        Stream stream,
        (Func<StreamReader, TArg, TResult> converter, TArg arg, Encoding charset) args
      ) => args.converter(new StreamReader(stream, args.charset), args.arg);

      public TResult ReadAs<TArg, TResult>(
        Func<StreamReader, TArg, TResult> converter,
        TArg arg,
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBody(
        messageStreamConverter: ReadAsConverter,
        arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg, charset ?? DefaultCharset),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task<TResult> ReadAsAsync<TArg, TResult>(
        Func<StreamReader, TArg, Task<TResult>> converter,
        TArg arg,
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadMessageBodyAsync(
        messageStreamConverter: ReadAsConverter,
        arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg, charset ?? DefaultCharset),
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      /*
       * ReadAsString()
       */
      private static string ReadAsStringConverter(StreamReader reader) => reader.ReadToEnd();
      private static Task<string> ReadAsStringAsyncConverter(StreamReader reader) => reader.ReadToEndAsync();

      public string ReadAsString(
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadAs(
        converter: ReadAsStringConverter,
        charset: charset,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task<string> ReadAsStringAsync(
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadAsAsync(
        converter: ReadAsStringAsyncConverter,
        charset: charset,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      /*
       * ReadAsStringLines()
       */
      private static IReadOnlyList<string> ReadAsStringLinesConverter(StreamReader reader) => reader.ReadAllLines();
      private static Task<IReadOnlyList<string>> ReadAsStringLinesAsyncConverter(StreamReader reader) => reader.ReadAllLinesAsync();

      public IReadOnlyList<string> ReadAsStringLines(
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadAs(
        converter: ReadAsStringLinesConverter,
        charset: charset,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );

      public Task<IReadOnlyList<string>> ReadAsStringLinesAsync(
        Encoding charset = null,
        int maxLines = EntireMessage,
        CancellationToken cancellationToken = default,
        IProgress<PopMessageRetrieveProgressInfo> progress = null
      ) =>
      ReadAsAsync(
        converter: ReadAsStringLinesAsyncConverter,
        charset: charset,
        maxLines: maxLines,
        cancellationToken: cancellationToken,
        progress: progress
      );
    }
  }
}
