// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.Text;

namespace Smdn.Net.Pop3.Protocol.Transmission {
#if SERIALIZATION
  [Serializable]
#endif
  public class PopCommandResult
#if SERIALIZATION
    : ISerializable
#endif
  {

#if SERIALIZATION
    [Serializable]
#endif
    private sealed class SuccessCommandResult : PopCommandResult {
      public override bool IsConcludedByClientSide => true;
      public override bool IsSuccess => true;

      internal SuccessCommandResult(string description)
        : base(string.Concat("<", description, ">"))
      {
      }

#if SERIALIZATION
      internal SuccessCommandResult(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
      }
#endif
    }

#if SERIALIZATION
    [Serializable]
#endif
    private sealed class FailureCommandResult : PopCommandResult {
      public override bool IsConcludedByClientSide => true;
      public override bool IsSuccess => false;

      internal FailureCommandResult(string description)
        : base(string.Concat("<", description, ">"))
      {
      }

#if SERIALIZATION
      internal FailureCommandResult(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
      }
#endif
    }

    internal static PopCommandResult CreateSuccess(string description) => new SuccessCommandResult(description);
    internal static PopCommandResult CreateFailure(string description) => new FailureCommandResult(description);

    /*
     * instance members
     */
    public virtual bool IsConcludedByClientSide => StatusResponse == null;

    public virtual bool IsSuccess {
      get {
        if (StatusResponse == null)
          return false;

        if (StatusResponse.Status == PopStatusIndicator.Positive)
          return true;
        else
          return false;
      }
    }

    public string Description { get; }
    public PopStatusResponse StatusResponse { get; }
    public IReadOnlyList<PopResponse> ReceivedResponses { get; }

    private protected PopCommandResult(string description)
    {
      Description = description;
      ReceivedResponses =
#if NET45
        ArrayExtensions.Empty<PopResponse>();
#else
        Array.Empty<PopResponse>();
#endif
    }

    internal PopCommandResult(PopStatusResponse statusResponse, IReadOnlyList<PopResponse> receivedResponses)
    {
      Description = string.Concat("\"", statusResponse.ResponseText.GetTextAsString(), "\"");
      StatusResponse = statusResponse;
      ReceivedResponses = receivedResponses;
    }

#if SERIALIZATION
    internal protected PopCommandResult(SerializationInfo info, StreamingContext context)
    {
      Description = info.GetString(nameof(Description));
      StatusResponse = (PopStatusResponse)info.GetValue(nameof(StatusResponse), typeof(PopStatusResponse));
      ReceivedResponses = (PopResponse[])info.GetValue(nameof(ReceivedResponses), typeof(PopResponse[]));
    }

    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue(nameof(Description), Description);
      info.AddValue(nameof(StatusResponse), StatusResponse);
      info.AddValue(nameof(ReceivedResponses), ReceivedResponses.ToArray());
    }
#endif

    public bool IsResponseStatus(PopStatusIndicator status)
    {
      if (StatusResponse == null)
        return false;

      return StatusResponse.Status == status;
    }

    public PopStatusResponse GetResponseCode(PopResponseCode code)
    {
      foreach (var resp in ReceivedResponses) {
        var status = resp as PopStatusResponse;

        if (status != null && status.ResponseText.Code == code)
          return status;
      }

      return null;
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(IsSuccess), IsSuccess),
        (nameof(Description), Description),
        (nameof(StatusResponse), StatusResponse),
      });
    }
  }
}
