using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Threading;

using Smdn.Text.Encodings;

namespace Smdn.Net.Pop3 {
  public class PopPseudoServer : PseudoServer {
    private class Response : ResponseBase {
      public bool IsQuitResponse { get; set; }
      public bool IsPushResponse { get; set; }

      public Response(string expectedCommand,
                      string responseText,
                      Action actionBeforeRespond = null,
                      Action actionAfterRespond = null,
                      TimeSpan responseDelay = default)
        : base(expectedCommand: expectedCommand,
               responseText: responseText,
               actionBeforeRespond: actionBeforeRespond,
               actionAfterRespond: actionAfterRespond,
               responseDelay: responseDelay)
      {
      }
    }

    public PopPseudoServer()
      : base()
    {
    }

    protected override void Initialize()
    {
      base.Initialize();

      SetGreetingResponse("+OK\r\n");
    }

    public static void Start(Action<PopPseudoServer> action,
                             [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                             [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                             [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      Start(true, action, callerMemberName, sourceFilePath, sourceLinerNumber);
    }

    public static void Start(bool throwIfServerNotStopped,
                             Action<PopPseudoServer> action,
                             [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                             [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                             [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      using (var server = new PopPseudoServer()) {
        server.Start(true, throwIfServerNotStopped, action, callerMemberName, sourceFilePath, sourceLinerNumber);
      }
    }

    public static void StartNotAccepting(Action<PopPseudoServer> action,
                                         [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                                         [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                                         [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      using (var server = new PopPseudoServer()) {
        server.Start(false, false, action, callerMemberName, sourceFilePath, sourceLinerNumber);
      }
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond = null,
                                    Action actionAfterRespond = null,
                                    bool isQuitResponse = false)
    {
      if (greeting == null) {
        greetingResponse = null;
      }
      else {
        greetingResponse = new Response(expectedCommand: null,
                                        responseText: greeting,
                                        actionBeforeRespond: actionBeforeRespond,
                                        actionAfterRespond: actionAfterRespond);

        greetingResponse.IsQuitResponse = isQuitResponse;
      }
    }

    public void EnqueuePushResponse(string responseText)
    {
      InternalEnqueueResponse(expectedCommand: null,
                              responseText: responseText,
                              actionBeforeRespond: null,
                              actionAfterRespond: null,
                              isQuitResponse: false,
                              isPushResponse: true,
                              responseDelay: default);
    }

    public void EnqueueResponse(string responseText)
    {
      InternalEnqueueResponse(expectedCommand: null,
                              responseText: responseText,
                              actionBeforeRespond: null,
                              actionAfterRespond: null,
                              isQuitResponse: false,
                              isPushResponse: false,
                              responseDelay: default);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond = null,
                                Action actionAfterRespond = null,
                                bool isQuitResponse = false,
                                TimeSpan responseDelay = default)
    {
      InternalEnqueueResponse(expectedCommand: expectedCommand,
                              responseText: responseText,
                              actionBeforeRespond: actionBeforeRespond,
                              actionAfterRespond: actionAfterRespond,
                              isQuitResponse: isQuitResponse,
                              isPushResponse: false,
                              responseDelay: responseDelay);
    }

    private void InternalEnqueueResponse(string expectedCommand,
                                         string responseText,
                                         Action actionBeforeRespond,
                                         Action actionAfterRespond,
                                         bool isQuitResponse,
                                         bool isPushResponse,
                                         TimeSpan responseDelay)
    {
      var response = new Response(expectedCommand: expectedCommand,
                                  responseText: responseText,
                                  actionBeforeRespond: actionBeforeRespond,
                                  actionAfterRespond: actionAfterRespond,
                                  responseDelay: responseDelay);

      response.IsQuitResponse = isQuitResponse;
      response.IsPushResponse = isPushResponse;

      responseQueue.Enqueue(response);
    }

    // TODO: rename
    public string DequeueRequest()
    {
      lock (commandQueue) {
        if (0 < commandQueue.Count)
          return commandQueue.Dequeue();
        else
          return null;
      }
    }

    protected override void ServerProc(TcpListener server, TcpClient client)
    {
      var commandStream = client.GetStream();

      if (greetingResponse != null) {
        Send(greetingResponse, commandStream);

        if (greetingResponse.IsQuitResponse)
          return;
      }

      var commandBuffer = new byte[0x1000];

      for (;;) {
        if (!IsConnected(client))
          return;

        Response pushResponse = null;

        lock (responseQueue) {
          if (0 < responseQueue.Count && responseQueue.Peek().IsPushResponse)
            pushResponse = responseQueue.Dequeue();
        }

        if (pushResponse != null) {
          Send(pushResponse, commandStream);

          if (pushResponse.IsQuitResponse)
            return;
          else
            continue;
        }

        /*        
         * wait for command sent
         */
        string commandText = null;

        try {
          var commandLength = commandStream.Read(commandBuffer, 0, commandBuffer.Length);

          if (commandLength == 0 && !IsConnected(client))
            return;

          commandText = OctetEncoding.EightBits.GetString(commandBuffer, 0, commandLength);

          WriteLog("C:", commandText);

          lock (commandQueue) {
            commandQueue.Enqueue(commandText);
          }
        }
        catch (IOException ex) {
          if (!client.Connected)
            return; // disconnected

          var exSocket = ex.InnerException as SocketException;

          if (exSocket != null &&
              (exSocket.SocketErrorCode == SocketError.WouldBlock || exSocket.SocketErrorCode == SocketError.TimedOut)) {
            continue;
          }

          throw;
        }

        /*
         * send response
         */
        while (responseQueue.Count == 0) {
          SetClientRequestReceived();

          if (!IsConnected(client))
            return;

          Thread.Sleep(10);
        }

        var response = responseQueue.Dequeue();

        Send(response, commandStream, commandText);

        if (response.IsQuitResponse)
          return;
      }
    }

    private readonly Queue<Response> responseQueue = new Queue<Response>();
    private readonly Queue<string> commandQueue = new Queue<string>();
    private Response greetingResponse;
  }
}



