using System;
using System.Net;
using NUnit.Framework;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientProfileTests {
    [Test]
    public void TestConstructor_Defaults()
    {
      var profileDefault = new PopClientProfile();
      var profile = new PopClientProfile("host");

      Assert.AreEqual("host", profile.Host);

      Assert.AreEqual(profileDefault.Port, profile.Port);
      Assert.AreEqual(profileDefault.SecurePort, profile.SecurePort);
      Assert.AreEqual(profileDefault.UserName, profile.UserName);
      Assert.AreEqual(profileDefault.AuthType, profile.AuthType);
      Assert.AreEqual(profileDefault.Timeout, profile.Timeout);
      Assert.AreEqual(profileDefault.SendTimeout, profile.SendTimeout);
      Assert.AreEqual(profileDefault.ReceiveTimeout, profile.ReceiveTimeout);
      Assert.AreEqual(profileDefault.UseTlsIfAvailable, profile.UseTlsIfAvailable);
      CollectionAssert.AreEqual(profileDefault.UsingSaslMechanisms, profile.UsingSaslMechanisms);
      Assert.AreEqual(profileDefault.AllowInsecureLogin, profile.AllowInsecureLogin);
      Assert.AreEqual(profileDefault.AllowStandardLogin, profile.AllowStandardLogin);
      Assert.AreEqual(profileDefault.AllowApopLogin, profile.AllowApopLogin);

      CollectionAssert.Contains(profile.UsingSaslMechanisms, "CRAM-MD5");
      CollectionAssert.Contains(profile.UsingSaslMechanisms, "DIGEST-MD5");
    }

    [Test] public void TestHost_Null() => Host_ArgumentException<ArgumentNullException>(null);
    [Test] public void TestHost_Empty() => Host_ArgumentException<ArgumentException>(string.Empty);

    private void Host_ArgumentException<TExpectedException>(string hostValue)
      where TExpectedException : ArgumentException
    {
      var ex1 = Assert.Throws<TExpectedException>(() => new PopClientProfile(hostValue));

      Assert.AreEqual(nameof(PopClientProfile.Host), ex1.ParamName);

      var profile = new PopClientProfile();

      var ex2 = Assert.Throws<TExpectedException>(() => profile.Host = hostValue);

      Assert.AreEqual(nameof(PopClientProfile.Host), ex2.ParamName);
    }

    [Test]
    public void TestClone()
    {
      var original = new PopClientProfile(new Uri("pop://user@localhost/"));

      original.Timeout = 1;
      original.ReceiveTimeout = 2;
      original.SendTimeout = 3;
      original.UseTlsIfAvailable = false;
      original.SetUsingSaslMechanisms(new[] { "X-SASL-EXT" });
      original.AllowInsecureLogin = true;
      original.AllowStandardLogin = false;
      original.AllowApopLogin = false;

      var cloned = original.Clone();

      Assert.AreNotSame(cloned, original);
      Assert.AreNotSame(original.Authority, cloned.Authority);
      Assert.AreEqual(original.Authority.AbsoluteUri, cloned.Authority.AbsoluteUri);
      Assert.AreEqual(1, cloned.Timeout);
      Assert.AreEqual(2, cloned.ReceiveTimeout);
      Assert.AreEqual(3, cloned.SendTimeout);
      Assert.IsFalse(cloned.UseTlsIfAvailable);
      Assert.AreNotSame(original.UsingSaslMechanisms, cloned.UsingSaslMechanisms);
      CollectionAssert.AreEqual(original.UsingSaslMechanisms, cloned.UsingSaslMechanisms);
      Assert.IsTrue(cloned.AllowInsecureLogin);
      Assert.IsFalse(cloned.AllowStandardLogin);
      Assert.IsFalse(cloned.AllowApopLogin);
    }

    [Test]
    public void TestCloneUsingSaslMechanismsEmpty()
    {
      var original = new PopClientProfile(new Uri("pop://user@localhost/"));

      original.UsingSaslMechanisms.Clear();

      var cloned = original.Clone();

      Assert.AreNotSame(cloned, original);
      Assert.AreNotSame(original.Authority, cloned.Authority);
      Assert.AreEqual(original.Authority.AbsoluteUri, cloned.Authority.AbsoluteUri);
      Assert.IsEmpty(cloned.UsingSaslMechanisms);
    }

    [Test]
    public void TestSerializeBinary()
    {
      var original = new PopClientProfile(new Uri("pop://user@localhost/"));

      original.Timeout = 1;
      original.ReceiveTimeout = 2;
      original.SendTimeout = 3;
      original.UseTlsIfAvailable = false;
      original.SetUsingSaslMechanisms(new[] { "X-SASL-EXT" });
      original.AllowInsecureLogin = true;
      original.AllowStandardLogin = false;
      original.AllowApopLogin = false;

      TestUtils.Assert.IsSerializableBinaryFormat(original, deserialized => {
        Assert.AreEqual(original.Authority.AbsoluteUri, deserialized.Authority.AbsoluteUri);
        Assert.AreEqual(1, deserialized.Timeout);
        Assert.AreEqual(2, deserialized.ReceiveTimeout);
        Assert.AreEqual(3, deserialized.SendTimeout);
        Assert.IsFalse(deserialized.UseTlsIfAvailable);
        Assert.AreNotSame(original.UsingSaslMechanisms, deserialized.UsingSaslMechanisms);
        CollectionAssert.AreEqual(original.UsingSaslMechanisms, deserialized.UsingSaslMechanisms);
        Assert.IsTrue(deserialized.AllowInsecureLogin);
        Assert.IsFalse(deserialized.AllowStandardLogin);
        Assert.IsFalse(deserialized.AllowApopLogin);
      });
    }

    [Test]
    public void TestSerializeBinary_UsingSaslMechanismsEmpty()
    {
      var original = new PopClientProfile(new Uri("pop://user@localhost/"));

      original.UsingSaslMechanisms.Clear();

      TestUtils.Assert.IsSerializableBinaryFormat(original, deserialized => {
        Assert.AreEqual(original.Authority.AbsoluteUri, deserialized.Authority.AbsoluteUri);
        Assert.IsEmpty(deserialized.UsingSaslMechanisms);
      });
    }
  }
}
