using System;
using System.IO;
using NUnit.Framework;

using Smdn.Text.Encodings;

namespace Smdn.Net.Pop3.Protocol.Transmission {
  [TestFixture]
  public class PopResponseRecieverTests {
    private PopResponse ReceiveResponse(string responseLine)
    {
      var receiver = new PopPseudoResponseReceiver();

      receiver.SetResponse(responseLine);

      return receiver.ReceiveResponse();
    }

    [TestCase("+OK []\r\n")]
    [TestCase("+OK [] \r\n")]
    [TestCase("+OK [] text\r\n")]
    public void TestMalformed_RespText_Empty(string resp)
    {
      Assert.Throws<PopMalformedResponseException>(() => ReceiveResponse(resp));
    }

    [Test]
    public void TestStatusOkResponse()
    {
      var response = ReceiveResponse("+OK\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Positive, status.Status);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestStatusOkResponseWithText()
    {
      var response = ReceiveResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Positive, status.Status);
      Assert.AreEqual("POP3 server ready <1896.697170952@dbc.mtview.ca.us>", status.Text);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestStatusNoResponse()
    {
      var response = ReceiveResponse("-ERR\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Negative, status.Status);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestStatusNoResponseWithText()
    {
      var response = ReceiveResponse("-ERR sorry, no mailbox for frated here\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Negative, status.Status);
      Assert.IsNull(status.ResponseText.Code);
      Assert.AreEqual("sorry, no mailbox for frated here", status.Text);
    }

    [Test]
    public void TestStatusNoResponseWithResponseCode()
    {
      var response = ReceiveResponse("-ERR [IN-USE] Do you have another POP session running?\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Negative, status.Status);
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.AreEqual(PopResponseCode.InUse, status.ResponseText.Code);
      Assert.AreEqual("Do you have another POP session running?", status.Text);
    }

    [Test]
    public void TestStatusNoResponseWithResponseCodeOnly()
    {
      var response = ReceiveResponse("-ERR [IN-USE]\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Negative, status.Status);
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.AreEqual(PopResponseCode.InUse, status.ResponseText.Code);
      Assert.IsEmpty(status.Text);
    }

    [Test]
    public void TestStatusOkResponseWithExtraSpace()
    {
      var response = ReceiveResponse("+OK \r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Positive, status.Status);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestStatusNoResponseWithExtraSpace()
    {
      var response = ReceiveResponse("-ERR \r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Negative, status.Status);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestStatusResponseWithResponseCodeAndTextWithoutSpace()
    {
      var response = ReceiveResponse("-ERR [IN-USE]text\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopStatusResponse, "response type");

      var status = response as PopStatusResponse;

      Assert.AreEqual(PopStatusIndicator.Negative, status.Status);
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.AreEqual(PopResponseCode.InUse, status.ResponseText.Code);
      Assert.AreEqual("text", status.Text);
    }

    [Test]
    public void TestCommandContinuationRequest()
    {
      var response = ReceiveResponse("+\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopContinuationRequest, "response type");
    }

    [Test]
    public void TestCommandContinuationRequestWithBase64Text()
    {
      var response = ReceiveResponse("+ TlRMTVNTUAACAAAAAAAAAAAAAAABAoAAfvE6HR6OGggAAAAAAAAAABQAFAAwAAAAAwAMAGgAYQB5AGEAdABlAAAAAAA=\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopContinuationRequest, "response type");

      var continuation = response as PopContinuationRequest;

      Assert.AreEqual("TlRMTVNTUAACAAAAAAAAAAAAAAABAoAAfvE6HR6OGggAAAAAAAAAABQAFAAwAAAAAwAMAGgAYQB5AGEAdABlAAAAAAA=", continuation.Base64Text.ToString());
    }

    [Test]
    public void TestTerminationResponse()
    {
      var response = ReceiveResponse(".\r\n");

      Assert.IsNotNull(response);
      Assert.IsTrue(response is PopTerminationResponse, "response type");
    }

    [Test]
    public void TestHandleAsMultiline()
    {
      var lines = new[] {
        "+OK 134 octets",
        "MIME-Version: 1.0",
        "Content-Type: text/plain",
        "",
        "...byte-stuffed",
        "+OK response style line",
        "-ERR response style line2",
        "+ continuation request style line",
        "+",
        "..",
        "end of message",
        ".",
        string.Empty,
      };

      var receiver = new PopPseudoResponseReceiver();

      receiver.SetResponse(string.Join("\r\n", lines));

      var expectedResponseTypes = new[] {
        typeof(PopStatusResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopFollowingResponse),
        typeof(PopTerminationResponse),
      };

      for (var i = 0; i < expectedResponseTypes.Length; i++) {
        var response = receiver.ReceiveResponse();

        Assert.IsNotNull(response);
        Assert.IsInstanceOf(expectedResponseTypes[i], response, "response #{0}", i);

        if (response is PopFollowingResponse) {
          var expected = lines[i];

          if (expected.StartsWith("."))
            expected = expected.Substring(1);

          Assert.AreEqual((response as PopFollowingResponse).Text.ToString(), expected);
        }

        if (i == 0)
          receiver.HandleAsMultiline = true;
      }
    }

    [Test]
    public void TestReceiveLine()
    {
      var lines = new[] {
        "+OK 134 octets",
        "MIME-Version: 1.0",
        "Content-Type: text/plain",
        "",
        "...byte-stuffed",
        "+OK response style line",
        "-ERR response style line2",
        "+ continuation request style line",
        "+",
        "..",
        "end of message",
        ".",
        string.Empty,
      };

      var receiver = new PopPseudoResponseReceiver();

      receiver.SetResponse(string.Join("\r\n", lines));

      for (var i = 0; i < lines.Length; i++) {
        var expected = lines[i] + "\r\n";

        if (expected.StartsWith("."))
          expected = expected.Substring(1);

        var lineStream = new MemoryStream();
        var ret = receiver.ReceiveLine(lineStream);

        lineStream.Dispose();

        var line = lineStream.ToArray();

        if (i == lines.Length - 2) {
          Assert.IsFalse(ret);
          break;
        }
        else {
          Assert.IsTrue(ret);
          Assert.AreEqual(expected, OctetEncoding.SevenBits.GetString(line));
        }
      }
    }
  }
}
